'''
This file is from: https://raw.githubusercontent.com/bearpaw/pytorch-classification/master/models/cifar/resnet.py
by Wei Yang
'''
import torch
import torch.nn.functional as F
import torch.nn as nn
import math

__all__ = ['resnet']

def conv3x3(in_planes, out_planes, stride=1):
    "3x3 convolution with padding"
    return nn.Conv2d(in_planes, out_planes, kernel_size=3, stride=stride,
                     padding=1, bias=False)

class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, downsample=None, relu_at_the_end: bool = True):
        super(BasicBlock, self).__init__()
        self.conv1 = conv3x3(inplanes, planes, stride)
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = conv3x3(planes, planes)
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride
        self.relu_at_the_end = relu_at_the_end
        
    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        # might comment out residual adding part for identity init
        out += residual
        
        if self.relu_at_the_end:
            out = self.relu(out)

        return out

# NOT USING BOTTLENECK
class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, inplanes, planes, stride=1, downsample=None):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(inplanes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride,
                               padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, planes * 4, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(planes * 4)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out

class Mask_Model_Resnet(nn.Module):

    def __init__(self, mask_output, depth, block_name='BasicBlock'):
        super(Mask_Model_Resnet, self).__init__()
        
        self.mask_output = mask_output
        
        # add a temperature parameter
        
        
        # Model type specifies number of layers for CIFAR-10 model
        if block_name.lower() == 'basicblock':
            assert (depth - 2) % 6 == 0, 'When use basicblock, depth should be 6n+2, e.g. 20, 32, 44, 56, 110, 1202'
            n = (depth - 2) // 6
            block = BasicBlock
        elif block_name.lower() == 'bottleneck':
            assert (depth - 2) % 9 == 0, 'When use bottleneck, depth should be 9n+2, e.g. 20, 29, 47, 56, 110, 1199'
            n = (depth - 2) // 9
            block = Bottleneck
        else:
            # raise ValueError('block_name shoule be Basicblock or Bottleneck')
            raise ValueError('block_name shoule be Basicblock')

        self.inplanes = 16
        self.conv1 = nn.Conv2d(3, 16, kernel_size=3, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(16)
        self.relu = nn.ReLU(inplace=True)
        
        # self.layer1 = self._make_layer(block, 8, n, stride=1)
        # self.layer2 = self._make_layer(block, 4, n, stride=1) # we have changed stride from 2 to 1 to keep the size intact;
        # self.layer3 = self._make_layer(block, 1, n, stride=1) # we have changed stride from 2 to 1 to keep the size intact;

        if mask_output == "penalty":
            relu_at_the_end = True
        else:
            relu_at_the_end = False

        self.layer1 = self._make_layer(block, 16, n, stride=2, relu_at_the_end=relu_at_the_end)
        self.layer2 = self._make_layer(block, 32, n, stride=1, relu_at_the_end=relu_at_the_end)
        self.layer3 = self._make_layer(block, 1, n, stride=1, relu_at_the_end=relu_at_the_end)

        # self.avgpool = nn.AvgPool2d(8)
        # self.fc = nn.Linear(64 * block.expansion, num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

        # final layer depending on mask output mechanism
        if self.mask_output == "sigmoid":
            self.sigmoid_layer = nn.Sigmoid()
        elif self.mask_output == "penalty":
            self.final_conv_layer = nn.Conv2d(1, 1, kernel_size=1, stride=1)                

    def _make_layer(self, block, planes, blocks, stride=1, relu_at_the_end: bool = True):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(self.inplanes, planes * block.expansion,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, relu_at_the_end))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, relu_at_the_end=relu_at_the_end))

        return nn.Sequential(*layers)

    def forward(self, x):
        
        original_size = x.size()[2:]
        
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        # x = self.layer1(x)
        # x = self.layer2(x)
        # x = self.layer3(x)

        # x = self.layer1(x)
        # x = self.layer2(x) 
        # x = self.layer3(x) 

        x = self.layer1(x)
        # print("x shape after layer1 {}".format(x.shape))
        x = self.layer2(x)
        # print("x shape after layer2 {}".format(x.shape))
        x = self.layer3(x)
        # print("x shape after layer3 {}".format(x.shape))

        # x = F.upsample(x, original_size, mode='bilinear') --> deprecated by Pytorch
        x = F.interpolate(x, size=original_size, mode='bilinear')

        if self.mask_output == "sigmoid":
            x = self.sigmoid_layer(x)
        elif self.mask_output == "penalty":
            x = self.final_conv_layer(x)

        # won't use the below since we want to output masks of same shape as input
        # x = self.avgpool(x)
        # x = x.view(x.size(0), -1)
        # x = self.fc(x)

        # return torch.squeeze(x)
        return x

def mask_model_resnet(**kwargs):
    return Mask_Model_Resnet(**kwargs)